#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""This module contains functions and classes to create lists of sorted
file/path names and compressed tar archives from such lists.

The sorting algorithm tries to arrange the given file names in a way to
increase the "compressability" by grouping files with same extensions so
files with similar content structures end up close to each other in the
archive file.

:todo: Maybe change blocking factor default to 1.
"""
from __future__ import division

#
# TODO Switch from printing to `sys.stderr` to `logging` the output.
# 
__author__ = "Marc 'BlackJack' Rintsch"
__version__ = '0.4.0'
__date__ = '2014-04-26'

__docformat__ = "reStructuredText"

import os
import sys
import tarfile

from itertools import chain, imap, izip

try:
    from lzma import LZMACompressor
    HAVE_LZMA = True
except ImportError:
    HAVE_LZMA = False
#:
#: Maps compression algorithm name to short version.
#:
ALGORITHM_TO_SHORT_NAME = {
    'none': '', 'gzip': 'gz', 'bzip2': 'bz2'
}
if HAVE_LZMA:
    ALGORITHM_TO_SHORT_NAME['lzma'] = 'xz'


class LZMAWriter(object):
    """Wraps a file like object and transparently compresses the written data
    with the LZMA algorithm.
    """
    def __init__(self, out_file):
        """The given `out_file` must be a file like object, opened for writing
        in binary mode.
        """
        self._out_file = out_file
        self._compressor = LZMACompressor()

    def __enter__(self):
        return self

    def __exit__(self, *_args):
        self.close()

    def write(self, data):
        self._out_file.write(self._compressor.compress(data))

    def close(self):
        self._out_file.write(self._compressor.flush())
        self._out_file.close()


class FileName(object):
    """A `FileName` represents a path and provides quick access to
    some components like the file name and its extension.
    
    Beside those attributes some test functions to see what kind of
    file belongs to this `FileName` from the `os.path` module are
    provided as methods on `FileName` objects.
    
    :ivar fullpath: normalized path name.
    :ivar path: path part of `fullpath`.
    :ivar name: file name part of `fullpath`.
    :ivar extension: file name extension in lower case.
    :type fullpath path name extension: str
    
    :see: `os.path`
    """
    def __init__(self, path):
        self.fullpath = os.path.normpath(path)
        self.path, self.name = os.path.split(self.fullpath)
        self.extension = os.path.splitext(self.name)[1].lower()
    
    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.fullpath)
    
    def __str__(self):
        """Returns the normalized version of the path given to the
        constructor.
        """
        return self.fullpath
    
    def __len__(self):
        return len(str(self))

    def __cmp__(self, other):
        """Compares two `FileName` objects.
        
        The comparison is done by file name first and path as second
        sorting key, i.e. files with the same names but different paths
        are "grouped" together by sorting `FileName` objects.
        """
        return cmp(self.name, other.name) or cmp(self.path, other.path)
        # TODO: Sort by file size if name and path are equal!?
    
    def isfile(self):
        """Tests if `FileName` refers to a file (or link to a file)."""
        return os.path.isfile(self.fullpath)
    
    def isdir(self):
        """Tests if `FileName` refers to a directory (or link to a
        directory).
        """
        return os.path.isdir(self.fullpath)
    
    def isspecial(self):
        """Tests if `FileName` is "special", i.e. not a regular file
        or directory.
        """
        return not ((self.isfile() or self.isdir())
                    and not os.path.islink(self.fullpath))


class FileList(object):
    """A container with `FileName` elements and the sorting logic.
    
    A `FileList` offers several ways to add elements and is iterable.
    
    When iterating over the container the elements are returned sorted
    in a way that (should) make the sequence of files more "compressable".
    This is done by grouping files with same file name extensions
    together as those files are more likely to contain redundant
    information.  See `FileList.__iter__()` for more information about
    the sorting order.
    
    :ivar _directories: directories in this container.
    :type _directories: list
    :ivar _groups: lists of `FileName` objects grouped by file name extension.
    :type _groups: dict
    :ivar _specials: special files in this container
                     (see `FileName.isspecial()`).
    :type _specials: list
    
    :see: `FileName`
    """
    def __init__(self, filenames=()):
        """Creates a `FileList`.
        
        :param filenames: file names to populate this container with.
        :type filenames: iterable of `str` or `FileName` objects.
        """
        self._directories = list()
        self._groups = dict()
        self._specials = list()
        self.add_files(filenames)
    
    def __iter__(self):
        """Creates iterator over the sorted elements in this container.
        
        How the elements are sorted
        ---------------------------
        
        First of all there are three groups:
        
        1. Directories,
        2. regular files,
        3. and "special" files, i.e. everything not in group 1 or 2.
        
        The regular files are grouped by their file name extension and
        every group is sorted by file name.
        
        :returns: iterator over the sorted elements in this container.
        :rtype: iterable
        """
        self._directories.sort()
        for directory in self._directories:
            yield directory

        groups = self._groups.items()
        groups.sort()
        for (dummy, paths) in groups:
            paths.sort()
            for path in paths:
                yield path

        self._specials.sort()
        for special in self._specials:
            yield special
    
    def __len__(self):
        """Returns the number of elements in this container."""
        return (sum(map(len, self._groups.itervalues()))
                + len(self._directories)
                + len(self._specials))
        
    def add_file(self, filename):
        """Adds an element to this container.
        
        :param filename: the file name to add.
        :type filename: str or `FileName`
        """
        if isinstance(filename, str):
            filename = FileName(filename)
        
        if filename.isspecial():
            self._specials.append(filename)
        elif filename.isdir():
            self._directories.append(filename)
        else:
            group = self._groups.setdefault(filename.extension, list())
            group.append(filename)
        
    def add_files(self, filenames):
        """Adds elements to this container.
        
        :param filenames: file names to add.
        :type filenames: iterable of `str` or `FileName` objects.
        """
        for filename in filenames:
            self.add_file(filename)


def iter_filenames(path):
    """Iterates recursivly over the contents of `path`.
    
    The function does not follow links but returns them.
    
    :param path: name of a file or directory.
    :type path: `str`
    
    :returns: an iterator over all file names in the directory
        and its subdirectories or just an iterator with the
        `FileName` of `path` if `path` was not a directory.
    :rtype: iterable of `FileName` objects
    
    :bug: The function silently ignores errors.
    """
    yield FileName(path)
    
    if os.path.isdir(path):
        # FIXME: Silently ignores errors.
        for (dirpath, dirnames, filenames) in os.walk(path):
            for name in chain(dirnames, filenames):
                yield FileName(os.path.join(dirpath, name))


def compress(fileobj, paths, algorithm='bz2', path_count=None):
    """Compresses a list of files in ``tar`` format into an open file
    object.
    
    The `fileobj` can be anything with a ``write()`` method.
    
    :param fileobj: the file like object the data is written to.
    :type fileobj: `file` like
    :param paths: the paths of the files that will be tarred and compressed.
    :type paths: iterable of `str`
    :param algorithm: the compression algorithm.  Currently one of '',
        'gz', or 'bz2'.
    :type algorithm: str
    :param path_count: number of paths in `paths`.  Used to calculate and
        display how many percent of files are already processed.  If `None`
        and the length of `paths` cannot be determined the output is
        suppressed.
    :type path_count: int
    
    :raises CompressionError: if `algorithm` is not supported.
    """
    mode = "w|" + algorithm
    tar = tarfile.open(mode=mode, fileobj=fileobj, bufsize=tarfile.RECORDSIZE)
    
    try:
        path_count = len(paths)
    except TypeError:
        pass
    
    for i, path in enumerate(paths, 1):
        if path_count is not None:
            print >> sys.stderr, '[%3d%%]' % ((i / path_count) * 100),
        print >> sys.stderr, path
        try:
            tar.add(path, recursive=False)
        except OSError, error:
            print >> sys.stderr, error
    
    tar.close()


def _create_parser():
    """Creates an `optparse.OptionParser` and populates it with options.
    
    :returns: the parser.
    :rtype: `optparse.OptionParser`
    
    :todo: Use `optparse.OptionGroup` to group the options.
    :todo: Maybe use a default of 1 for the ``-b`` option.
    """
    # 
    # TODO Switch to `argparse`.
    # 
    from optparse import OptionParser
    
    usage = "usage: %prog [options] file(s)"
    
    parser = OptionParser(usage=usage, version='%prog ' + __version__)
    #
    # File name options.
    #
    parser.add_option("-o", "-f", "--file", dest="filename", action="store",
                      type="string", default=None,
                      help="write archive to this file instead of STDOUT.")
    parser.add_option('-a', '--auto-name', dest='auto_name',
                      action='store_true', default=False,
                      help=('infer archive file name from the first given '
                            'directory name.  This only works if there is '
                            'just one directory name given as argument.  '
                            'The archive is named: '
                            '``<directory_name>.tar[.<algorithm>]``'))
    parser.add_option('--list', dest='just_list', action='store_true',
                      default=False,
                      help=("just dump the sorted file names to STDOUT -- "
                            "don't create archive."))
    #
    # Compression options.
    #
    compression_algorithms = sorted(ALGORITHM_TO_SHORT_NAME)
    parser.add_option(
        '--compression', dest='algorithm', action='store', type='choice',
        choices=compression_algorithms, default='bzip2',
        help=(
            'select compression algorithm from {0}, or {1}. [%default]'
                .format(
                    ', '.join(compression_algorithms[:-1]),
                    compression_algorithms[-1]
                )
        )
    )
    parser.add_option('-z', '--gzip', dest='algorithm', action='store_const',
                      const='gzip',
                      help='use gzip compression.')
    parser.add_option('-j', '--bzip2', dest='algorithm', action='store_const',
                      const='bzip2',
                      help='use bzip2 compression. [default]')
    if 'lzma' in ALGORITHM_TO_SHORT_NAME:
        parser.add_option(
            '-J', '--lzma', dest='algorithm', action='store_const',
            const='lzma', help='use LZMA compression.'
        )
    # 
    # Experimental options.
    # 
    parser.add_option('-b', '--blocking-factor', dest='blocking_factor',
                      metavar='BLOCKS', action='store', type='int', default=20,
                      help='BLOCKS x 512 per record [%default]')

    return parser


def main():
    """Main function."""
    parser = _create_parser()
    (options, args) = parser.parse_args()
    
    if len(args) < 1:
        parser.print_help()
        sys.exit(1)

    # 
    # Set the (experimental) blocking factor.
    # 
    tarfile.RECORDSIZE = options.blocking_factor * tarfile.BLOCKSIZE
    
    options.algorithm = ALGORITHM_TO_SHORT_NAME[options.algorithm]
    
    if options.filename and options.auto_name:
        parser.error('Options -o/-f/--file and -a are mutually exclusive.')

    if options.auto_name:
        if len(args) > 1 or not os.path.isdir(args[0]):
            parser.error("Autonaming (-a) doesn't work if there is more than "
                         "one argument and if the one is no directory.")
        else:
            extension = os.path.extsep + 'tar'
            if options.algorithm:
                extension += os.path.extsep + options.algorithm
            options.filename = os.path.normpath(args[0]) + extension

    #
    # Create file name list.
    #
    print >> sys.stderr, "creating file list...",
    file_list = FileList(chain(*map(iter_filenames, args)))
    file_count = len(file_list)
    print >> sys.stderr, file_count, 'file(s) found.'

    #
    # Just list names or compress files.
    #
    if options.just_list:
        for name in file_list:
            print name
    else:
        if options.filename:
            outfile = open(options.filename, "wb")
        else:
            outfile = sys.stdout
        
        if options.algorithm == 'xz':
            outfile = LZMAWriter(outfile)
            options.algorithm = ''

        with outfile:
            compress(
                outfile, imap(str, file_list), options.algorithm, file_count
            )


if __name__ == '__main__':
    main() 
